/**
 * Created by Steam engine corp in 08.07.2009 23:33:54
 *
 * [kisser]xmlstreamer
 * @author Christopher Marlowe
 */

package ru.steamengine.xmlstreaming;

import org.jdom.Element;
import org.jdom.output.Format;
import org.jdom.output.XMLOutputter;
import ru.steamengine.rtti.basetypes.*;
import static ru.steamengine.rtti.defaultimplementors.utils.CommonStreamingUtils.*;
import ru.steamengine.streaming.basetypes.*;
import ru.steamengine.streaming.defaultimplementors.DefaultObjectNamespace;
import static ru.steamengine.xmlstreaming.XMLReadWriteConsts.*;

import java.io.OutputStream;
import java.util.ArrayList;
import java.util.List;


public class XMLObjectWriter implements ObjectWriter, StreamInit {


    private static final Object[] NO_OBJECTS = new Object[0];

    private boolean inited = false;

    private Element rootElement;

    private Object root;

    private RegistryUser registryUser;

    private final ObjectNamespace namespace = new DefaultObjectNamespace();

    private ObjectResolver objectResolver;

    public XMLObjectWriter() {
    }

    public XMLObjectWriter(RegistryUser registryUser) {
        this();
        initialize(registryUser);
    }

    public void write(final Object root, OutputStream stream) throws WriteError {
        try {
            startWriting();
            this.root = root;
            registryUser.doTransaction(new RegistryUser.Transaction() {
                @Override
                public void inTransaction() {
                    writeObject(-1, root, "");
                }
            });

            if (rootElement != null) {
                XMLOutputter outputter = new XMLOutputter(Format.getPrettyFormat());
                outputter.output(rootElement, stream);
            }
        } catch (Exception e) {
            throw new WriteError(e);
        }
    }

    @Override
    public void write(final Object root, ObjectResolver objectResolver, OutputStream stream) throws WriteError {
        try {
            startWriting();
            this.root = root;
            this.objectResolver = objectResolver;
            registryUser.doTransaction(new RegistryUser.Transaction() {
                @Override
                public void inTransaction() {
                    writeObject(-1, root, "");
                }
            });

            if (rootElement != null) {
                XMLOutputter outputter = new XMLOutputter(Format.getPrettyFormat());
                outputter.output(rootElement, stream);
            }
        } catch (Exception e) {
            throw new WriteError(e);
        }
    }

    public Object getRoot() {
        return root;
    }


    private void checkInited() {
        if (!inited)
            throw new IllegalArgumentException("object is not inited");
    }


    private void writeObjectProps(final Object obj,
                                  final List<Element> propList,
                                  final String propPath) {

        RTTIEntries rtti = registryUser.getRTTI(obj.getClass());
        for (RTTIEntry entry : rtti.getProperties()) {
            PropertyTypeIdent ident = getPropTypeIdent(entry, registryUser);
            Class propClass = entry.getPropType();
            String propertyName = entry.getPropName();
            String newPath = getPropPath(propPath, propertyName);

            switch (ident) {
                case value:
                    boolean valStored = entry.isStored(obj) && !entry.isDefault(obj);
                    if (valStored) {
                        Element valElement = new Element(PROP_IDENT);
                        try {
                            final Object val = entry.getVal(obj);

                            PropertyRecoder r = registryUser.getSimpleType(propClass);
                            if (r == null)
                                throw new IllegalArgumentException("recoder is null");
                            if (val == null && (r instanceof PropertyRecoderEx) && !((PropertyRecoderEx) r).isNullable())
                                throw new IllegalArgumentException("value for " +
                                        propClass.getName() +
                                        " can not be null");
                            valElement.setAttribute(PATH_IDENT, newPath);

                            if (val != null) {
                                @SuppressWarnings({"unchecked"})
                                byte[] bytes = r.getValue(val);
                                String valValue = new String(bytes, UsedCharset.CHARSET);
                                valElement.setAttribute(VALUE_IDENT, valValue);
                            }

                        } finally {
                            propList.add(valElement);
                        }
                    }


                    break;

                case nested:
                    Object nestedObject = entry.getVal(obj);
                    if (nestedObject != null) {
                        writeObjectProps(nestedObject, propList, newPath);
                    }
                    break;


                case reference:
                    boolean refStored = entry.isStored(obj);
                    if (refStored) {
                        Object ref = entry.getVal(obj);
                        final Element refElement = new Element(PROP_IDENT);
                        propList.add(refElement);
                        if (ref != null) {
                            String classIdent = registryUser.findClassIdent(ref.getClass());
                            String refValue = getObjectName(ref, classIdent);
                            refElement.setAttribute(PATH_IDENT, newPath);
                            refElement.setAttribute(VALUE_IDENT, refValue);
                        } else {// null is written
                            refElement.setAttribute(PATH_IDENT, newPath);
                        }
                    }
                    break;


                default:
                    throw new IllegalArgumentException("");

            }


        }

        // processing a list
        for (ListSlot slot : rtti.getLists()) {
            String propertyName = slot.getPropertyName();
            String newPath = getPropPath(propPath, propertyName);
//          Class itemClass = listedItem.getItemClass();
            Class itemClass = slot.getItemClass();
            Element listElements = new Element(PROP_IDENT);

            listElements.setAttribute(PATH_IDENT, newPath);
            PropertyRecoder st = registryUser.getSimpleType(itemClass);
            int count = 0;
            if (st != null) {
                // it is a simple type list
                for (Object o : getListIterator(slot, obj)) {
                    if ((o == null) && (st instanceof PropertyRecoderEx) && !((PropertyRecoderEx) st).isNullable())
                        throw new WriteError("value can not be null");

                    Element element = new Element(ITEM_IDENT);
                    listElements.addContent(element);
                    if (o != null) {
                        //noinspection unchecked
                        byte[] bytes = st.getValue(o);
                        String valAsString = new String(bytes, UsedCharset.CHARSET);
//                        element.setAttribute(PATH_IDENT, newPath);
                        element.setAttribute(VALUE_IDENT, valAsString);
                    } else {
//                        element.setAttribute(PATH_IDENT, newPath);
                    }
                    count++;
                }

            } else {
                for (Object o : getListIterator(slot, obj)) {
                    Element listedElement = writeListedObject(o, "");
                    listElements.addContent(listedElement);
                    count++;
                }
            }

            if (count > 0)
                propList.add(listElements);


        }
    }

    private Element writeObject(final int position,
                                final Object obj,
                                final String propPath) {
        if (propPath == null)
            throw new NullPointerException("propPath is null");


        Element element = new Element(OBJECT_IDENT);
        String classIdent = registryUser.getClassIdent(obj.getClass());
        element.setAttribute(CLASS_IDENT, classIdent);
        element.setAttribute(NAME_IDENT, getObjectName(obj, classIdent));
        if (position >= 0)
            element.setAttribute(POSITION_IDENT, String.valueOf(position));
        if (rootElement == null)
            rootElement = element;

        List<Element> elements = new ArrayList<Element>();
        writeObjectProps(obj, elements, propPath); // object properties are collected here

        ObjectProcessor processor = registryUser.getObjProcessor(obj.getClass());
        CountedIterator iterator = getObjectIterator(processor, obj);
        while (iterator.hasNext()) {
            int index = iterator.nextIndex();
            Object child = iterator.next();
            Element el = writeObject(index, child, propPath);
            elements.add(el);
        }

        element.setContent(elements);
        return element;
    }


    /**
     * @param obj    an object
     * @param prefix prefix
     * @return is what is written to the stream
     */
    private String getObjectName(Object obj, String prefix) {
        prefix = prefix == null ? "obj" : prefix;
        String name;
        ObjectNameProcessor nameProcessor = registryUser.getNameProcessor(obj.getClass());
        if (objectResolver != null && ((name = objectResolver.getObjectName(obj)) != null)) {
            return namespace.setObjectName(obj, name);
        } else if (nameProcessor == null) {
            return namespace.getObjectName(obj, prefix);
        } else if ((name = nameProcessor.getName(obj)) == null) {
            return namespace.getObjectName(obj, prefix);
        } else if (!namespace.isNameUnique(obj, name)) {
            return namespace.getObjectName(obj, prefix);
        } else {
            return namespace.setObjectName(obj, name);
        }
    }


    private Element writeListedObject(final Object obj,
                                      final String propPath) {
        if (propPath == null)
            throw new NullPointerException("propPath is null");

        Element element = new Element(ITEM_IDENT);
        List<Element> elements = new ArrayList<Element>();
        writeObjectProps(obj, elements, propPath); // object props are collected here

        element.setContent(elements);
        return element;
    }

    private void startWriting() {
        rootElement = null;
        namespace.clear();
    }

    @Override
    public <T> T initialize(RegistryUser registryUser) {
        if (inited)
            throw new IllegalArgumentException("object is already inited");

        if (registryUser == null)
            throw new NullPointerException("registryUser is null");

        this.registryUser = registryUser;
        this.inited = true;

        //noinspection unchecked
        return (T) this;
    }

}